
using Azure.Messaging.EventHubs;
using Azure.Messaging.EventHubs.Producer;
using Azure.Storage.Blobs;

using Microsoft.Azure.WebJobs;
using Microsoft.Extensions.Logging;

using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Schema;

using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

using TemplateContent.Common;
using TemplateContent.Common.Models;

namespace TemplateContent.IngressProcessor
{
    public static class IngressProcessorController
    {
        private static ILogger _logger;

        [FunctionName("IngressProcessor")]
        public static async Task Run(
            [EventHubTrigger("%afBindingInputEventHubName%", Connection = "afBindingInputEventHubConnectionString")] EventData iotHubMessage,
            ILogger logger)
        {
            if (iotHubMessage == null)
                throw new ArgumentNullException(nameof(iotHubMessage));

            if (logger == null)
                throw new ArgumentNullException(nameof(logger));

            _logger = logger;

            string messageContent = Encoding.UTF8.GetString(iotHubMessage.EventBody);
            string deviceId = iotHubMessage.SystemProperties["iothub-connection-device-id"].ToString();

            _logger.LogDebug($"SCHEMA CONTROL for message-deviceID:{deviceId}-Message:{messageContent}");

            JObject jIoTHubMessage = JObject.Parse(messageContent);

            try
            {
                if (jIoTHubMessage.TryGetValue("$schema", out JToken schemaPath))
                {
                    //Message schema control
                    string schemaContent = GetSchemaContent(schemaPath.Value<string>());

                    var result = CheckSchemaMessage(messageContent, schemaContent);

                    if (result != null && result.IsMessageValid)
                    {
                        _logger.LogDebug($"   >>>SCHEMA OK");

                        //Add commissioning control at this level
                        string? checkCommissioning = Environment.GetEnvironmentVariable("checkCommissioning");
                        if (bool.TryParse(checkCommissioning, out bool checkCommissioningBool) && checkCommissioningBool)
                        {
                            _logger.LogDebug("   >>>COMMISSIONING CHECK");

                            if (await IsDeviceCommissioned(deviceId))
                                await ProcessOutputs(iotHubMessage, deviceId);
                            else
                                _logger.LogWarning($"   >>>DEVICE {deviceId} NOT COMMISSIONED");

                        }
                        else
                        {
                            _logger.LogDebug("   >>>NO COMMISSIONING CHECK");
                            await ProcessOutputs(iotHubMessage, deviceId);
                        }
                    }
                    else
                    {
                        string errorMessage = String.Empty;
                        if (result.Errors != null)
                            errorMessage = string.Join(Environment.NewLine, result.Errors);

                        _logger.LogDebug($"   >>>SCHEMA KO::{deviceId}::Message::{Environment.NewLine}{messageContent}::{Environment.NewLine}Errors::{Environment.NewLine}{errorMessage}");

                        if (bool.Parse(Environment.GetEnvironmentVariable("afEnableSchemaErrorsStorage")))
                            await SchemaErrorsStorageHelper.ProcessSchemaErrors(_logger, result, deviceId, String.Empty, "telemetry");
                    }
                }
                else
                {
                    _logger.LogDebug($"   >>>NO SCHEMA PROVIDED.THE MESSAGE CANNOT BE ACCEPTED.");
                }
            }
            catch (Exception ex)
            {
                _logger.LogError(ex.Message);
            }
        }

        #region Private methods
        private static async Task ProcessOutputs(EventData iotHubMessage, string deviceId)
        {
            Parallel.Invoke(
                                    async () =>
                                    {
                                        _logger.LogDebug("   >>>PARALLEL PROCESS: SENDING THE MESSAGE TO THE EVENT HUB (OUTPUT)");

                                        //EVENT HUB
                                        string eventHubConnectionString = Environment.GetEnvironmentVariable("afBindingOutputEventHubConnectionString", EnvironmentVariableTarget.Process);
                                        string eventHubName = Environment.GetEnvironmentVariable("afBindingOutputEventHubName", EnvironmentVariableTarget.Process);
                                        await using (var producer = new EventHubProducerClient(eventHubConnectionString, eventHubName))
                                        {
                                            using EventDataBatch eventBatch = await producer.CreateBatchAsync();
                                            eventBatch.TryAdd(iotHubMessage);

                                            await producer.SendAsync(eventBatch);
                                        }
                                    },
                                    async () =>
                                    {
                                        _logger.LogDebug("   >>>PARALLEL PROCESS: SENDING THE MESSAGE TO THE STORAGE (OUTPUT)");

                                        //STORAGE                                        
                                        //WARNING: default Blob output binding does not support using the deviceId in the path.
                                        string storageConnectionString = Environment.GetEnvironmentVariable("afBindingOutputStorageConnectionString", EnvironmentVariableTarget.Process);
                                        string storageContainerName = Environment.GetEnvironmentVariable("afBindingValidatedTelemetryStoragePath", EnvironmentVariableTarget.Process);
                                        DateTime currentDate = DateTime.UtcNow;

                                        BlobServiceClient blobServiceClient = new BlobServiceClient(storageConnectionString);
                                        BlobContainerClient containerClient = blobServiceClient.GetBlobContainerClient(storageContainerName);
                                        BlobClient blobClient = containerClient.GetBlobClient($"{deviceId}/{currentDate.ToString("yyyy-MM-dd")}/{currentDate.ToString("yyyy-MM-ddTHH:mm:ssZ")}.json");

                                        //iotHubMessage.ContentType = "application/json";
                                        iotHubMessage.Properties.Add("validatedTelemetry", "true");

                                        await blobClient.UploadAsync(iotHubMessage.EventBody, true);
                                    });
        }

        private static async Task<bool> IsDeviceCommissioned(string deviceId)
        {
            bool result = false;

            try
            {
                using (HttpClient client = new HttpClient())
                {
                    string checkCommissioningEndpointUrl = Environment.GetEnvironmentVariable("checkCommissioningEndpointUrl", EnvironmentVariableTarget.Process);
                    string responseBody = await client.GetStringAsync(new Uri($"{checkCommissioningEndpointUrl}/{deviceId}"));

                    bool.TryParse(responseBody, out result);
                }
            }
            catch (HttpRequestException e)
            {
                _logger.LogError($"IsDeviceCommissioned::{e.Message}");
            }

            return result;
        }

        /// <summary>
        /// Gets a blob from Azure Blob storage. Those blobs should contain JSON schemas.
        /// </summary>
        /// <param name="schemaContentPath"></param>
        /// <returns></returns>
        private static string GetSchemaContent(string schemaContentPath)
        {
            if (string.IsNullOrEmpty(schemaContentPath))
                throw new ArgumentNullException(nameof(schemaContentPath));

            string result = string.Empty;

            BlobClient blobClient = new BlobClient(new Uri(schemaContentPath));
            var schemaBlobDownloadResponse = blobClient.Download();

            if (schemaBlobDownloadResponse != null && schemaBlobDownloadResponse.Value != null)
            {
                using (var downLoadInfo = schemaBlobDownloadResponse.Value)
                using (var streamedContent = new StreamReader(downLoadInfo.Content))
                {
                    result = streamedContent.ReadToEnd();
                }
            }

            return result;
        }

        /// <summary>
        /// Checks if a JSON message respects a given JSON Schema.
        /// </summary>
        /// <param name="message"></param>
        /// <param name="schema"></param>
        /// <returns></returns>
        private static SchemaValidationResponse CheckSchemaMessage(string message, string schema)
        {
            if (string.IsNullOrEmpty(message))
                throw new ArgumentNullException(nameof(message));

            if (string.IsNullOrEmpty(schema))
                throw new ArgumentNullException(nameof(schema));

            SchemaValidationResponse result = new SchemaValidationResponse { Message = message, IsMessageValid = false };

            try
            {

                JSchemaUrlResolver resolver = new JSchemaUrlResolver();
                JSchema jSchema = JSchema.Parse(schema, resolver);
                JObject jsonMessage = JObject.Parse(message);

                IList<string> errors = null;
                if (!jsonMessage.IsValid(jSchema, out errors))
                {
                    if (errors != null && errors.Count > 0)
                    {
                        result.Errors = errors;
                    }
                }
                else
                {
                    result.IsMessageValid = true;
                }
            }
            catch (Exception ex)
            {
                _logger.LogError($"CheckSchema - Error {ex.Message}");
            }


            return result;
        }
        #endregion
    }
}
